import argparse
import logging
import shutil
import sys

from weir import zfs


log = logging.getLogger(__name__)

# copy zfs snapshots using send and receive
def copy(src, dst, base=None, intermediates=False,
		replicate=False, properties=False, deduplicate=False,
		append_path=False, append_name=False, force=False, nomount=False):
	with src.send(base=base, intermediates=intermediates, replicate=replicate,
			properties=properties, deduplicate=deduplicate) as fsrc:
		with zfs.receive(dst, append_path=append_path, append_name=append_name,
				force=force, nomount=nomount) as fdst:
			shutil.copyfileobj(fsrc, fdst, length=131072)

def main(argv=sys.argv):
	# set up command-line options
	parser = argparse.ArgumentParser()
	parser.add_argument('-v', '--verbose', action='count', default=0)
#	parser.add_argument('-n', '--dry-run', action='store_true')
	parser.add_argument('-F', '--force', action='store_true')

	# stream options
	parser.add_argument('-R', '--replicate', action='store_true')
	parser.add_argument('-D', '--deduplicate', action='store_true')
	parser.add_argument('-p', '--properties', action='store_true')

	# destination options
	append_args = parser.add_mutually_exclusive_group()
	append_args.add_argument('-d', '--append-path', action='store_true')
	append_args.add_argument('-e', '--append-name', action='store_true')
	parser.add_argument('-u', '--no-mount', action='store_true')

	# incremental base
	base_args = parser.add_mutually_exclusive_group()
	base_args.add_argument('-i', metavar='base', dest='inc_base')
	base_args.add_argument('-I', metavar='base', dest='int_base')

	# source and destination
	parser.add_argument('src')
	parser.add_argument('dst')

	# parse command-line args
	try:
		args = parser.parse_args(argv[1:])
	except SystemExit:
		return 2

	# set log level - use WARNING by default
	log_level = max(logging.WARNING - (args.verbose * 10), logging.DEBUG)
	logging.basicConfig(level=log_level)

	# interpret incremental snapshot options
	intermediates = (args.int_base is not None)
	base = args.int_base if intermediates else args.inc_base

	# copy datasets
	try:
		src = zfs.open(args.src)
		copy(src, args.dst, base, intermediates,
			args.replicate, args.properties, args.deduplicate,
			args.append_path, args.append_name, args.force, args.no_mount)
		return 0
	except Exception as exc:
		log.error(exc)
		return 1

if __name__ == '__main__':
	sys.exit(main())



from __future__ import division

import argparse
import logging
import re
import sys

from datetime import datetime
from datetime import timedelta

from weir import process, zfs


log = logging.getLogger(__name__)

# implement timedelta.total_seconds() for Python 2.6 support
def total_seconds(td):
	return ((td.days * 24 * 3600 + td.seconds) * 10**6
		+ td.microseconds) / 10**6

# destroy expired snapshots
def cull(filesystem, min_age=None, max_age=None,
		max_density=None, dry_run=False):
	# configure exclusions
	exclusions = []
	# keep all snapshots under min-age
	if min_age is not None:
		exclusions.append(lambda s: s['age'] < min_age)

	# configure filters
	filters = []
	# destroy all snapshots over max-age
	if max_age is not None:
		filters.append(lambda s: s['age'] > max_age)
	# destroy snapshots exceeding max-density
	# where density is defined as age / delta
	if max_density is not None:
		filters.append(lambda s: total_seconds(s['age']) \
			/ total_seconds(s['delta']) > max_density)

	# get snapshots as list of name, creation-time pairs
	# XXX: weir should provide values of correct type
	snapshots = [(prop['name'], datetime.fromtimestamp(float(prop['value'])))
		for prop in zfs.findprops(
			filesystem, max_depth=1, props=('creation',), types=('snapshot',))]

	# use most recent backup to calculate snapshot age
	latest = snapshots[-1][1]
	# use previous snapshot to calculate delta
	prev = None

	# always keep most recent snapshot
	for name, ctime in snapshots[:-1]:
		# create snapshot object to test against filters
		s = {
			'name': name,
			'ctime': ctime,
			'age': latest - ctime,
			'delta': abs(ctime - prev)
				if prev is not None else timedelta.max,
		}

		# skip if snapshot not eligible for destruction
		if (any(f(s) for f in exclusions)
				or not any(f(s) for f in filters)):
			prev = ctime
			continue

		# destroy snapshot
		log.info('destroying %s', name)
		if not dry_run:
			# ignore errors, eg due to held snapshot
			try:
				snapshot = zfs.ZFSSnapshot(name)
				snapshot.destroy()
			except process.DatasetBusyError:
				pass

# parse iso8601 time duration
def parse_timedelta(s):
	pattern = r'''^p?(?:(\d+)y)?(?:(\d+)m)?(?:(\d+)w)?(?:(\d+)d)?
		t?(?:(\d+)h)?(?:(\d+)m)?(?:(\d+)s)?$'''
	match = re.match(pattern, s, re.IGNORECASE|re.VERBOSE)
	if not match:
		raise ValueError(s)

	y, m, w, d, H, M, S = (int(x or 0) for x in match.groups())

	# approximate 1 year = 365.25 days and 1 month = 1/12 year
	d = y * 365.25 + m * 365.25 / 12 + w * 7 + d

	return timedelta(days=d, hours=H, minutes=M, seconds=S)

def main(argv=sys.argv):
	# set up command-line options
	parser = argparse.ArgumentParser()
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('-n', '--dry-run', action='store_true')
	parser.add_argument('--min-age', type=parse_timedelta)
	parser.add_argument('--max-age', type=parse_timedelta)
	parser.add_argument('--max-density', type=float)
	parser.add_argument('filesystem')

	# parse command-line args
	try:
		args = parser.parse_args(argv[1:])
	except SystemExit:
		return 2

	# set log level - use WARNING by default
	log_level = max(logging.WARNING - (args.verbose * 10), logging.DEBUG)
	logging.basicConfig(level=log_level)

	# cull filesystems
	try:
		fs = args.filesystem
		cull(fs, args.min_age, args.max_age, args.max_density, args.dry_run)
		return 0
	except Exception as exc:
		log.error(exc)
		return 1

if __name__ == '__main__':
	sys.exit(main())



import argparse
import logging
import sys

from weir import process, zfs

from sluice.copy import copy


log = logging.getLogger(__name__)

# synchronise snapshots on source and destination datasets
def sync(src, dst, intermediates=False, recursive=False, force=False):
	if recursive:
		raise NotImplementedError()

	# get all source snapshots
	is_snapshot = isinstance(src, zfs.ZFSSnapshot)
	src_snaps = src.parent().snapshots() if is_snapshot else src.snapshots()

	# find index of source snapshot if specified
	n = len(src_snaps) - 1
	if is_snapshot:
		while n >= 0:
			if src_snaps[n].name == src.name: break
			n -= 1
		else:
			raise AssertionError("snapshot not found in parent's children")

	# source and destination snapshot names must match for sync
	if zfs.ZFSSnapshot(dst).snapname():
		raise TypeError('destination must not be a snapshot')

	# find earlier snapshots not present in destination
	try:
		dst_snaps = set(s.snapname() for s in zfs.open(dst).snapshots())
	except process.DatasetNotFoundError:
		m = -1
	else:
		m = n
		while m >= 0:
			if src_snaps[m].snapname() in dst_snaps: break
			m -= 1

	# send full snapshot if no common ancestor found
	# note: don't mount new filesystem to avoid need for -F later
	if m < 0 and n >= 0:
		m = 0 if intermediates else n
		copy(src_snaps[m], dst, force=force, nomount=True)

	# send incremental snapshots up to specified snapshot
	if intermediates >= 2:
		# legacy mode: send each snapshot separately
		for i in range(m+1, n+1):
			copy(src_snaps[i], dst, base=src_snaps[i-1].name, force=force)
	elif m < n:
		# normal mode: send one single- or multi-snapshot stream
		copy(src_snaps[n], dst, base=src_snaps[m].name,
			intermediates=intermediates, force=force)

	# update snapshot holds
	try:
		# tag src snapshots with destination dataset name
		tag = 'sluice.sync:' + dst

		# place hold on synced source snapshot
		if n >= 0:
			src_snaps[n].hold(tag)

		# release any previous holds
		for i in range(n):
			try:
				src_snaps[i].release(tag)
			except process.HoldTagNotFoundError:
				pass
	except process.HoldTagExistsError:
		pass

def main(argv=sys.argv):
	# set up command-line options
	parser = argparse.ArgumentParser()
	parser.add_argument('-v', '--verbose', action='count', default=0)
#	parser.add_argument('-n', '--dry-run', action='store_true')
	parser.add_argument('-F', '--force', action='store_true')

	parser.add_argument('-r', '--recursive', action='store_true')
	parser.add_argument('-I', '--intermediates', action='count', default=0)

	parser.add_argument('src')
	parser.add_argument('dst')

	# parse command-line args
	try:
		args = parser.parse_args(argv[1:])
	except SystemExit:
		return 2

	# set log level - use WARNING by default
	log_level = max(logging.WARNING - (args.verbose * 10), logging.DEBUG)
	logging.basicConfig(level=log_level)

	# sync source to destination
	try:
		src = zfs.open(args.src)
		sync(src, args.dst, intermediates=args.intermediates,
			recursive=args.recursive, force=args.force)
		return 0
	except Exception as exc:
		log.error(exc)
		return 1

if __name__ == '__main__':
	sys.exit(main())



import argparse
import logging
import sys

from datetime import datetime

from weir import zfs


log = logging.getLogger(__name__)

# ISO 8601 compliant default timestamp format
DEFAULT_FORMAT = '%Y-%m-%dT%H%M'

# zfs property holding the preferred timestamp format
FORMAT_PROPERTY = 'sluice.autosnapshot:format'

def autosnapshot(filesystem, format=None, recursive=False, props={}):
	if not format:
		format = filesystem.getpropval(FORMAT_PROPERTY, DEFAULT_FORMAT)

	snapname = datetime.now().strftime(format)

	log.info('creating new snapshot %s@%s', filesystem.name, snapname)
	return filesystem.snapshot(snapname, recursive, props)

def parse_prop(s):
	name, _, value = s.partition('=')
	return name, value

def main(argv=sys.argv):
	# set up command-line options
	parser = argparse.ArgumentParser()
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('-r', '--recursive', action='store_true')
	parser.add_argument('-o', metavar='property=value',
		dest='props', default=[], action='append', type=parse_prop)
	parser.add_argument('snapshot', metavar='filesystem[@snapformat]')

	# parse command-line args
	try:
		args = parser.parse_args(argv[1:])
	except SystemExit:
		return 2

	# set log level - use WARNING by default
	log_level = max(logging.WARNING - (args.verbose * 10), logging.DEBUG)
	logging.basicConfig(level=log_level)

	# create snapshot
	try:
		template = zfs.ZFSSnapshot(args.snapshot)
		dataset = template.parent()
		format = template.snapname()

		recursive = args.recursive
		props = dict(args.props)

		autosnapshot(dataset, format, recursive, props)
		return 0
	except Exception as exc:
		log.error(exc)
		return 1

if __name__ == '__main__':
	sys.exit(main())



from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

install_requires=[
    'weir>=0.4.0,<0.5.*',
]

# extra requirements for Python 2.6
try:
    import argparse
except ImportError:
    install_requires.append('argparse>=1.3.0')

setup(
    name='sluice',
    version='0.3.1',
    description='Tools for managing zfs snapshots',
    long_description=long_description,

    url='https://bitbucket.org/stevedrake/sluice',
    author='Stephen Drake',
    author_email='steve@synergyconsultingnz.com',

    license='Common Development and Distribution License (CDDL)',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
    keywords='zfs',

    packages=['sluice'],

    install_requires=install_requires,

    entry_points = {
        'console_scripts': [
            'zfs-autosnapshot = sluice.autosnapshot:main',
            'zfs-copy = sluice.copy:main',
            'zfs-cull = sluice.cull:main',
            'zfs-sync = sluice.sync:main',
        ],
    }
)
